@@ -59,13 +59,24 @@
 #include <sys/stat.h> /* for stat() */
 #endif
 
-
-
 #include "util.h"
 #include "sysutils.h"
 #include "i18n.h"
 #include "zb32.h"
 
+/* The name of the symbolic link to the file from which the process
+ * text was read.  */
+#if __linux__
+# define MYPROC_SELF_EXE "/proc/self/exe"
+#elif defined(__NetBSD__)
+# define MYPROC_SELF_EXE "/proc/curproc/exe"
+#elif defined(__illumos__) || defined(__sun)
+# define MYPROC_SELF_EXE "/proc/self/path/a.out"
+#else /* Assume other BSDs */
+# define MYPROC_SELF_EXE "/proc/curproc/file"
+#endif
+
+
 /* The GnuPG homedir.  This is only accessed by the functions
  * gnupg_homedir and gnupg_set_homedir.  Malloced.  */
 static char *the_gnupg_homedir;
@@ -90,7 +101,7 @@ static byte w32_portable_app;
 #endif /*HAVE_W32_SYSTEM*/
 
 #ifdef HAVE_W32_SYSTEM
-/* This flag is true if this process' binary has been installed under
+/* This flag is true if this process's binary has been installed under
    bin and not in the root directory as often used before GnuPG 2.1. */
 static byte w32_bin_is_bin;
 #endif /*HAVE_W32_SYSTEM*/
@@ -235,7 +246,7 @@ copy_dir_with_fixup (const char *newdir)
 /* Get the standard home directory.  In general this function should
    not be used as it does not consider a registry value (under W32) or
    the GNUPGHOME environment variable.  It is better to use
-   default_homedir(). */
+   gnupg_homedir(). */
 const char *
 standard_homedir (void)
 {
@@ -250,6 +261,7 @@ standard_homedir (void)
       if (w32_portable_app)
         {
           dir = xstrconcat (rdir, DIRSEP_S "home", NULL);
+          gpgrt_annotate_leaked_object (dir);
         }
       else
         {
@@ -261,6 +273,7 @@ standard_homedir (void)
             {
               dir = xstrconcat (path, "\\gnupg", NULL);
               xfree (path);
+              gpgrt_annotate_leaked_object (dir);
 
               /* Try to create the directory if it does not yet exists.  */
               if (gnupg_access (dir, F_OK))
@@ -278,7 +291,7 @@ standard_homedir (void)
 
 /* Set up the default home directory.  The usual --homedir option
    should be parsed later. */
-const char *
+static const char *
 default_homedir (void)
 {
   const char *dir;
@@ -338,7 +351,18 @@ default_homedir (void)
 
       p = copy_dir_with_fixup (dir);
       if (p)
-        dir = p;
+        {
+          /* A new buffer has been allocated with proper semantics.
+           * Assign this to DIR.  If DIR is passed again to
+           * copy_dir_with_fixup there will be no need for a fix up
+           * and the function returns NULL.  Thus we leak only once.
+           * Setting the homedir is usually a one-off task but might
+           * be called a second time.  We also ignore such extra leaks
+           * because we don't know who still references the former
+           * string.  */
+          gpgrt_annotate_leaked_object (p);
+          dir = p;
+        }
 
       if (!is_gnupg_default_homedir (dir))
         non_default_homedir = 1;
@@ -375,8 +399,10 @@ check_portable_app (const char *dir)
     }
   xfree (fname);
 }
+#endif /*HAVE_W32_SYSTEM*/
 
 
+#ifdef HAVE_W32_SYSTEM
 /* Determine the root directory of the gnupg installation on Windows.  */
 static const char *
 w32_rootdir (void)
@@ -428,7 +454,237 @@ w32_rootdir (void)
   /* Fallback to the hardwired value. */
   return GNUPG_LIBEXECDIR;
 }
+#endif /*HAVE_W32_SYSTEM*/
 
+
+#ifndef HAVE_W32_SYSTEM /* Unix */
+/* Determine the root directory of the gnupg installation on Unix.
+ * The standard case is that this function returns NULL so that the
+ * root directory as configured at build time is used.  However, it
+ * may return a static string with a different root directory, similar
+ * to what we do on Windows.  That second mode is triggered by the
+ * existence of a file gpgconf.ctl installed side-by-side to gpgconf.
+ * This file is parsed for keywords describing the actually to be used
+ * root directory.  There is no solid standard on Unix to locate the
+ * binary used to create the process, thus we support this currently
+ * only on Linux and BSD where we can look this info up using the proc
+ * file system.  If WANT_SYSCONFDIR is true the optional sysconfdir
+ * entry is returned.  */
+static const char *
+unix_rootdir (int want_sysconfdir)
+{
+  static int checked;
+  static char *dir;   /* for the rootdir  */
+  static char *sdir;  /* for the sysconfdir */
+
+  if (!checked)
+    {
+      char *p;
+      char *buffer;
+      size_t bufsize = 256-1;
+      int nread;
+      gpg_error_t err;
+      char *line;
+      size_t linelen;
+      ssize_t length;
+      estream_t fp;
+      char *rootdir;
+      char *sysconfdir;
+      const char *name;
+
+      for (;;)
+        {
+          buffer = xmalloc (bufsize+1);
+          nread = readlink (MYPROC_SELF_EXE, buffer, bufsize);
+          if (nread < 0)
+            {
+              err = gpg_error_from_syserror ();
+              buffer[0] = 0;
+              if ((name = getenv ("GNUPG_BUILD_ROOT")) && *name == '/')
+                {
+                  /* Try a fallback for systems w/o a supported /proc
+                   * file system if we are running a regression test.  */
+                  log_info ("error reading symlink '%s': %s\n",
+                            MYPROC_SELF_EXE, gpg_strerror (err));
+                  xfree  (buffer);
+                  buffer = xstrconcat (name, "/bin/gpgconf", NULL);
+                  log_info ("trying fallback '%s'\n", buffer);
+                }
+              break;
+            }
+          else if (nread < bufsize)
+            {
+              buffer[nread] = 0;
+              break;  /* Got it.  */
+            }
+          else if (bufsize >= 4095)
+            {
+              buffer[0] = 0;
+              log_info ("error reading symlink '%s': %s\n",
+                        MYPROC_SELF_EXE, "value too large");
+              break;
+            }
+          xfree (buffer);
+          bufsize += 256;
+        }
+      if (!*buffer)
+        {
+          xfree (buffer);
+          checked = 1;
+          return NULL;  /* Error - assume no gpgconf.ctl.  */
+        }
+
+      p = strrchr (buffer, '/');
+      if (!p)
+        {
+          xfree (buffer);
+          checked = 1;
+          return NULL;  /* Erroneous /proc - assume no gpgconf.ctl.  */
+        }
+      *p = 0;  /* BUFFER has the directory.  */
+      if ((p = strrchr (buffer, '/')))
+        {
+          /* Strip one part and expect the file below a bin dir.  */
+          *p = 0;
+          p = xstrconcat (buffer, "/bin/gpgconf.ctl", NULL);
+          xfree (buffer);
+          buffer = p;
+        }
+      else /* !p */
+        {
+          /* Installed in the root which is not a good idea.  Assume
+           * no gpgconf.ctl.  */
+          xfree (buffer);
+          checked = 1;
+          return NULL;
+        }
+
+      if (gnupg_access (buffer, F_OK))
+        {
+          /* No gpgconf.ctl file.  */
+          xfree (buffer);
+          checked = 1;
+          return NULL;
+        }
+      /* log_info ("detected '%s'\n", buffer); */
+      fp = es_fopen (buffer, "r");
+      if (!fp)
+        {
+          err = gpg_error_from_syserror ();
+          log_info ("error opening '%s': %s\n", buffer, gpg_strerror (err));
+          xfree (buffer);
+          checked = 1;
+          return NULL;
+        }
+
+      line = NULL;
+      linelen = 0;
+      rootdir = NULL;
+      sysconfdir = NULL;
+      while ((length = es_read_line (fp, &line, &linelen, NULL)) > 0)
+        {
+          /* Strip NL and CR, if present.  */
+          while (length > 0
+                 && (line[length - 1] == '\n' || line[length - 1] == '\r'))
+            line[--length] = 0;
+          trim_spaces (line);
+          if (!strncmp (line, "rootdir=", 8))
+            {
+              name = "rootdir";
+              p = line + 8;
+            }
+          else if (!strncmp (line, "rootdir =", 9))  /* (What a kludge) */
+            {
+              name = "rootdir";
+              p = line + 9;
+            }
+          else if (!strncmp (line, "sysconfdir=", 11))
+            {
+              name = "sysconfdir";
+              p = line + 11;
+            }
+          else if (!strncmp (line, "sysconfdir =", 12))  /* (What a kludge) */
+            {
+              name = "sysconfdir";
+              p = line + 12;
+            }
+          else
+            continue;
+          trim_spaces (p);
+          p = substitute_envvars (p);
+          if (!p)
+            {
+              err = gpg_error_from_syserror ();
+              log_info ("error getting %s from gpgconf.ctl: %s\n",
+                        name, gpg_strerror (err));
+            }
+          else if (!strcmp (name, "sysconfdir"))
+            {
+              xfree (sysconfdir);
+              sysconfdir = p;
+            }
+          else
+            {
+              xfree (rootdir);
+              rootdir = p;
+            }
+        }
+      if (es_ferror (fp))
+        {
+          err = gpg_error_from_syserror ();
+          log_info ("error reading '%s': %s\n", buffer, gpg_strerror (err));
+          es_fclose (fp);
+          xfree (buffer);
+          xfree (line);
+          xfree (rootdir);
+          xfree (sysconfdir);
+          checked = 1;
+          return NULL;
+        }
+      es_fclose (fp);
+      xfree (buffer);
+      xfree (line);
+
+      if (!rootdir || !*rootdir || *rootdir != '/')
+        {
+          log_info ("invalid rootdir '%s' specified in gpgconf.ctl\n", rootdir);
+          xfree (rootdir);
+          xfree (sysconfdir);
+          dir = NULL;
+        }
+      else if (sysconfdir && (!*sysconfdir || *sysconfdir != '/'))
+        {
+          log_info ("invalid sysconfdir '%s' specified in gpgconf.ctl\n",
+                    sysconfdir);
+          xfree (rootdir);
+          xfree (sysconfdir);
+          dir = NULL;
+        }
+      else
+        {
+          while (*rootdir && rootdir[strlen (rootdir)-1] == '/')
+            rootdir[strlen (rootdir)-1] = 0;
+          dir = rootdir;
+          gpgrt_annotate_leaked_object (dir);
+          /* log_info ("want rootdir '%s'\n", dir); */
+          if (sysconfdir)
+            {
+              while (*sysconfdir && sysconfdir[strlen (sysconfdir)-1] == '/')
+                sysconfdir[strlen (sysconfdir)-1] = 0;
+              sdir = sysconfdir;
+              gpgrt_annotate_leaked_object (sdir);
+              /* log_info ("want sysconfdir '%s'\n", sdir); */
+            }
+        }
+      checked = 1;
+    }
+
+  return want_sysconfdir? sdir : dir;
+}
+#endif /* Unix */
+
+
+#ifdef HAVE_W32_SYSTEM
 static const char *
 w32_commondir (void)
 {
@@ -460,6 +716,7 @@ w32_commondir (void)
            * version.  Use the installation directory instead.  */
           dir = xstrdup (rdir);
         }
+      gpgrt_annotate_leaked_object (dir);
     }
 
   return dir;
@@ -489,6 +746,10 @@ gnupg_set_homedir (const char *newdir)
   xfree (the_gnupg_homedir);
   the_gnupg_homedir = make_absfilename (newdir, NULL);;
   xfree (tmp);
+  /* Fixme: Should we use
+   *   gpgrt_annotate_leaked_object(the_gnupg_homedir)
+   * despite that we may free and allocate a new one in some
+   * cases?  */
 }
 
 
@@ -563,6 +824,7 @@ gnupg_daemon_rootdir (void)
         name = xstrdup ("/"); /* Error - use the curret top dir instead.  */
       else
         name = xstrdup (path);
+      gpgrt_annotate_leaked_object (name);
     }
 
   return name;
@@ -900,6 +1162,7 @@ gnupg_socketdir (void)
     {
       unsigned int dummy;
       name = _gnupg_socketdir_internal (0, &dummy);
+      gpgrt_annotate_leaked_object (name);
     }
 
   return name;
@@ -922,10 +1185,15 @@ gnupg_sysconfdir (void)
       s2 = DIRSEP_S "etc" DIRSEP_S "gnupg";
       name = xmalloc (strlen (s1) + strlen (s2) + 1);
       strcpy (stpcpy (name, s1), s2);
+      gpgrt_annotate_leaked_object (name);
     }
   return name;
 #else /*!HAVE_W32_SYSTEM*/
-  return GNUPG_SYSCONFDIR;
+  const char *dir = unix_rootdir (1);
+  if (dir)
+    return dir;
+  else
+    return GNUPG_SYSCONFDIR;
 #endif /*!HAVE_W32_SYSTEM*/
 }
 
@@ -933,28 +1201,35 @@ gnupg_sysconfdir (void)
 const char *
 gnupg_bindir (void)
 {
-#if defined (HAVE_W32CE_SYSTEM)
   static char *name;
-
-  if (!name)
-    name = xstrconcat (w32_rootdir (), DIRSEP_S "bin", NULL);
-  return name;
-#elif defined(HAVE_W32_SYSTEM)
   const char *rdir;
 
+#if defined(HAVE_W32_SYSTEM)
   rdir = w32_rootdir ();
   if (w32_bin_is_bin)
     {
-      static char *name;
-
       if (!name)
-        name = xstrconcat (rdir, DIRSEP_S "bin", NULL);
+        {
+          name = xstrconcat (rdir, DIRSEP_S "bin", NULL);
+          gpgrt_annotate_leaked_object (name);
+        }
       return name;
     }
   else
     return rdir;
 #else /*!HAVE_W32_SYSTEM*/
-  return GNUPG_BINDIR;
+  rdir = unix_rootdir (0);
+  if (rdir)
+    {
+      if (!name)
+        {
+          name = xstrconcat (rdir, DIRSEP_S "bin", NULL);
+          gpgrt_annotate_leaked_object (name);
+        }
+      return name;
+    }
+  else
+    return GNUPG_BINDIR;
 #endif /*!HAVE_W32_SYSTEM*/
 }
 
@@ -967,35 +1242,82 @@ gnupg_libexecdir (void)
 #ifdef HAVE_W32_SYSTEM
   return gnupg_bindir ();
 #else /*!HAVE_W32_SYSTEM*/
-  return GNUPG_LIBEXECDIR;
+  static char *name;
+  const char *rdir;
+
+  rdir = unix_rootdir (0);
+  if (rdir)
+    {
+      if (!name)
+        {
+          name = xstrconcat (rdir, DIRSEP_S "libexec", NULL);
+          gpgrt_annotate_leaked_object (name);
+        }
+      return name;
+    }
+  else
+    return GNUPG_LIBEXECDIR;
 #endif /*!HAVE_W32_SYSTEM*/
 }
 
 const char *
 gnupg_libdir (void)
 {
-#ifdef HAVE_W32_SYSTEM
   static char *name;
 
+#ifdef HAVE_W32_SYSTEM
   if (!name)
-    name = xstrconcat (w32_rootdir (), DIRSEP_S "lib" DIRSEP_S "gnupg", NULL);
+    {
+      name = xstrconcat (w32_rootdir (), DIRSEP_S "lib" DIRSEP_S "gnupg", NULL);
+      gpgrt_annotate_leaked_object (name);
+    }
   return name;
 #else /*!HAVE_W32_SYSTEM*/
-  return GNUPG_LIBDIR;
+  const char *rdir;
+
+  rdir = unix_rootdir (0);
+  if (rdir)
+    {
+      if (!name)
+        {
+          name = xstrconcat (rdir, DIRSEP_S "lib", DIRSEP_S, "gnupg", NULL);
+          gpgrt_annotate_leaked_object (name);
+        }
+      return name;
+    }
+  else
+    return GNUPG_LIBDIR;
 #endif /*!HAVE_W32_SYSTEM*/
 }
 
 const char *
 gnupg_datadir (void)
 {
-#ifdef HAVE_W32_SYSTEM
   static char *name;
 
+#ifdef HAVE_W32_SYSTEM
   if (!name)
-    name = xstrconcat (w32_rootdir (), DIRSEP_S "share" DIRSEP_S "gnupg", NULL);
+    {
+      name = xstrconcat (w32_rootdir (), DIRSEP_S "share" DIRSEP_S "gnupg",
+                         NULL);
+      gpgrt_annotate_leaked_object (name);
+    }
   return name;
 #else /*!HAVE_W32_SYSTEM*/
-  return GNUPG_DATADIR;
+  const char *rdir;
+
+  rdir = unix_rootdir (0);
+  if (rdir)
+    {
+      if (!name)
+        {
+          name = xstrconcat (rdir, DIRSEP_S "share" DIRSEP_S "gnupg", NULL);
+          gpgrt_annotate_leaked_object (name);
+        }
+      return name;
+    }
+  else
+    return GNUPG_DATADIR;
 #endif /*!HAVE_W32_SYSTEM*/
 }
 
@@ -1003,15 +1325,31 @@ gnupg_datadir (void)
 const char *
 gnupg_localedir (void)
 {
-#ifdef HAVE_W32_SYSTEM
   static char *name;
 
+#ifdef HAVE_W32_SYSTEM
   if (!name)
-    name = xstrconcat (w32_rootdir (), DIRSEP_S "share" DIRSEP_S "locale",
-                       NULL);
+    {
+      name = xstrconcat (w32_rootdir (), DIRSEP_S "share" DIRSEP_S "locale",
+                         NULL);
+      gpgrt_annotate_leaked_object (name);
+    }
   return name;
 #else /*!HAVE_W32_SYSTEM*/
-  return LOCALEDIR;
+  const char *rdir;
+
+  rdir = unix_rootdir (0);
+  if (rdir)
+    {
+      if (!name)
+        {
+          name = xstrconcat (rdir, DIRSEP_S "share" DIRSEP_S "locale", NULL);
+          gpgrt_annotate_leaked_object (name);
+        }
+      return name;
+    }
+  else
+    return LOCALEDIR;
 #endif /*!HAVE_W32_SYSTEM*/
 }
 
@@ -1030,7 +1368,6 @@ gpg_agent_socket_name (void)
   return name;
 }
 
-
 /* Return the user socket name used by DirMngr.  */
 const char *
 dirmngr_socket_name (void)
@@ -1038,7 +1375,25 @@ dirmngr_socket_name (void)
   static char *name;
 
   if (!name)
-    name = make_filename (gnupg_socketdir (), DIRMNGR_SOCK_NAME, NULL);
+    {
+      name = make_filename (gnupg_socketdir (), DIRMNGR_SOCK_NAME, NULL);
+      gpgrt_annotate_leaked_object (name);
+    }
+  return name;
+}
+
+
+/* Return the user socket name used by Keyboxd.  */
+const char *
+keyboxd_socket_name (void)
+{
+  static char *name;
+
+  if (!name)
+    {
+      name = make_filename (gnupg_socketdir (), KEYBOXD_SOCK_NAME, NULL);
+      gpgrt_annotate_leaked_object (name);
+    }
   return name;
 }
 
@@ -1099,6 +1454,8 @@ get_default_pinentry_name (int reset)
           else
             xfree (name2);
         }
+      if (name)
+        gpgrt_annotate_leaked_object (name);
     }
 
   return name;
@@ -1118,7 +1475,10 @@ static int gnupg_module_name_called;
  * be called before any invocation of 'gnupg_module_name', and must
  * not be called twice.  It can be used by test suites to make sure
  * the components from the build directory are used instead of
- * potentially outdated installed ones.  */
+ * potentially outdated installed ones.
+ * Fixme: It might be better to make use of the newer gpgconf.ctl feature
+ *        for regression testing.
+ */
 void
 gnupg_set_builddir (const char *newdir)
 {
@@ -1154,11 +1514,13 @@ gnupg_module_name (int which)
 
 #define X(a,b,c) do {                                                   \
     static char *name;                                                  \
-    if (!name)                                                          \
+    if (!name) {                                                        \
       name = gnupg_build_directory                                      \
         ? xstrconcat (gnupg_build_directory,                            \
                       DIRSEP_S b DIRSEP_S c EXEEXT_S, NULL)             \
         : xstrconcat (gnupg_ ## a (), DIRSEP_S c EXEEXT_S, NULL);       \
+      gpgrt_annotate_leaked_object (name);                              \
+    }                                                                   \
     return name;                                                        \
   } while (0)
 
@@ -1185,6 +1547,13 @@ gnupg_module_name (int which)
       X(libexecdir, "scd", "scdaemon");
 #endif
 
+    case GNUPG_MODULE_NAME_TPM2DAEMON:
+#ifdef GNUPG_DEFAULT_TPM2DAEMON
+      return GNUPG_DEFAULT_TPM2DAEMON;
+#else
+      X(libexecdir, "tpm2d", TPM2DAEMON_NAME);
+#endif
+
     case GNUPG_MODULE_NAME_DIRMNGR:
 #ifdef GNUPG_DEFAULT_DIRMNGR
       return GNUPG_DEFAULT_DIRMNGR;
@@ -1192,6 +1561,13 @@ gnupg_module_name (int which)
       X(bindir, "dirmngr", DIRMNGR_NAME);
 #endif
 
+    case GNUPG_MODULE_NAME_KEYBOXD:
+#ifdef GNUPG_DEFAULT_KEYBOXD
+      return GNUPG_DEFAULT_KEYBOXD;
+#else
+      X(libexecdir, "kbx", KEYBOXD_NAME);
+#endif
+
     case GNUPG_MODULE_NAME_PROTECT_TOOL:
 #ifdef GNUPG_DEFAULT_PROTECT_TOOL
       return GNUPG_DEFAULT_PROTECT_TOOL;
