@@ -224,6 +224,7 @@ struct
   unsigned int screen_lines;
   byte *show_subpackets;
   int rfc2440_text;
+  unsigned int min_rsa_length;   /* Used for compliance checks.  */
 
   /* If true, let write failures on the status-fd exit the process. */
   int exit_on_status_write_error;
@@ -245,6 +246,7 @@ struct
     unsigned int allow_old_cipher_algos:1;
     unsigned int allow_weak_digest_algos:1;
     unsigned int allow_weak_key_signatures:1;
+    unsigned int override_compliance_check:1;
     unsigned int large_rsa:1;
     unsigned int disable_signer_uid:1;
     unsigned int include_key_block:1;
