@@ -45,19 +45,6 @@
 #define MAX_DIGEST_LEN 64     /* Fits for SHA-512 */
 
 
-struct keyserver_spec
-{
-  struct keyserver_spec *next;
-
-  char *host;
-  int port;
-  char *user;
-  char *pass;
-  char *base;
-  unsigned int use_ldaps:1;
-};
-
-
 /* A large struct named "opt" to keep global flags. */
 EXTERN_UNLESS_MAIN_MODULE
 struct
@@ -155,7 +142,9 @@ struct
                                the integrity of the software at
                                runtime. */
 
-  struct keyserver_spec *keyserver;
+  unsigned int min_rsa_length;   /* Used for compliance checks.  */
+
+  strlist_t keyserver;
 
   /* A list of certificate extension OIDs which are ignored so that
      one can claim that a critical extension has been handled.  One
